import 'dart:async';
import 'dart:io';

import 'package:dio/dio.dart';
import 'package:get/get.dart';
import 'package:permission_handler/permission_handler.dart';
import 'package:rdc_flutter_sdk_2/rabbit_data.dart';
import 'package:rdc_flutter_sdk_2/src/controllers/condition_controller.dart';
import 'package:retry/retry.dart';
import 'package:rdc_flutter_sdk_2/src/constants/app_constants.dart' as con;
import 'package:rdc_flutter_sdk_2/src/controllers/device_info_controller.dart';
import 'package:rdc_flutter_sdk_2/src/controllers/location_controller.dart';
import 'package:rdc_flutter_sdk_2/src/controllers/rabbit_config_controller.dart';
import 'package:rdc_flutter_sdk_2/src/controllers/user_info_controller.dart';
import 'package:rdc_flutter_sdk_2/src/controllers/version_controller.dart';
import 'package:rdc_flutter_sdk_2/src/network/api_connector.dart';
import 'package:rdc_flutter_sdk_2/src/repositories/event_repo.dart';
import 'package:rxdart/subjects.dart';
import 'package:shared_preferences/shared_preferences.dart';
import 'package:pausable_timer/pausable_timer.dart';
import 'package:uuid/uuid.dart';

import '../util/date_time_helper.dart';

class EventController {
  /* ------------ service ------------*/
  late ApiConnector _api;
  /* ------------ service ------------*/

  /* ------------ repo ------------*/
  late EventRepo _eventRepo;
  /* ------------ repo ------------*/

  /* ------------ controller ------------*/
  final RabbitConfigController _config = Get.find(tag: 'rdc_flutter_sdk_2');
  final ConditionController _conditionController =
      Get.find(tag: 'rdc_flutter_sdk_2');
  final DeviceInfoController _deviceInfoController =
      Get.find(tag: 'rdc_flutter_sdk_2');
  final UserInfoController _userInfoController =
      Get.find(tag: 'rdc_flutter_sdk_2');
  final LocationController _locationController =
      Get.find(tag: 'rdc_flutter_sdk_2');
  final VersionController _versionController =
      Get.find(tag: 'rdc_flutter_sdk_2');
  /* ------------ controller ------------*/

  /* ------------ variable ------------*/
  final RabbitRequest rabbitRequest = RabbitRequest();
  var session = RabbitSession.empty;
  var _sessionId = '';
  DateTime _sessionStart = DateTime.now();
  int? loopTime;
  PausableTimer? sessionTask;
  PausableTimer? resumeTask;
  PausableTimer? rotationTask;

  var isLoading = false;
  var isFirstTime = true;
  /* ------------ variable ------------*/

  /* ------------ stream ------------*/
  var changeDB = BehaviorSubject.seeded('');
  var sizeDB = BehaviorSubject.seeded(0);
  var apiResult = BehaviorSubject.seeded('');
  /* ------------ stream ------------*/

  Future<EventController> init() async {
    _eventRepo = await EventRepo().init();
    _api = await ApiConnector().init();
    _initRequest();
    await checkOpenApp();
    await rotationLoop();
    return this;
  }

  Future<void> onChangeDebugMode() async {
    _api = await ApiConnector().init();
  }

  _initRequest() {
    rabbitRequest
      ..sdkVersion = _versionController.sdkVersion
      ..globalId = _deviceInfoController.globalId
      ..osType = _deviceInfoController.osType
      ..osVer = _deviceInfoController.osVersion
      ..display = _deviceInfoController.display
      ..maid = _deviceInfoController.maId
      ..device = _deviceInfoController.device
      ..model = _deviceInfoController.model
      ..brandName = _deviceInfoController.brandName
      ..board = _deviceInfoController.board;
  }

  Future<void> checkOpenApp() async {
    SharedPreferences sharedPreference = await SharedPreferences.getInstance();
    // check run app first time if session start -> set session end
    if (sharedPreference.containsKey('sessionId') &&
        sharedPreference.containsKey('pauseTime')) {
      var pauseTime = sharedPreference.getString('pauseTime')!;
      _sessionId = sharedPreference.getString('sessionId') ??
          const Uuid().v4().toString();
      session = RabbitSession.start;
      await sessionEnd(
          endTime: DateTime.parse(pauseTime).format(con.dateFormat));
    }

    if (_config.retentionDays > 0) {
      await _eventRepo.deleteAfterRetentionDays(_config.retentionDays);
    }

    isFirstTime = false;
  }

  Future<void> rotationLoop() async {
    if (_config.collectionType == DataCollectionType.rotation.name) {
      if (!isLoading) await uploadEvent();
      rotationTask?.cancel();
      rotationTask =
          PausableTimer(Duration(minutes: _config.rotationTime), () async {
        //print('rotationLoop : ${DateTime.now()}');
        await uploadEvent();
        rotationTask!
          ..reset()
          ..start();
      });
      rotationTask!.start();
    } else {
      rotationTask?.cancel();
    }
  }

  Future<void> sessionStart() async {
    SharedPreferences sharedPreference = await SharedPreferences.getInstance();
    session = RabbitSession.start;
    _sessionId = const Uuid().v4().toString();
    sharedPreference.setString('sessionId', _sessionId);
    _sessionStart = DateTime.now();
    sessionTask?.cancel();
    sessionTask =
        PausableTimer(Duration(minutes: _config.sessionExpire), () async {
      await _loopTime();
    });
    sessionTask!.start();
    await addEvent(
        eventName: 'session_start',
        eventValue: '',
        dateTime: _sessionStart.format(con.dateFormat));
  }

  Future<void> _loopTime() async {
    if (session == RabbitSession.start) {
      SharedPreferences sharedPreference =
          await SharedPreferences.getInstance();
      _sessionId = const Uuid().v4().toString();
      sharedPreference.setString('sessionId', _sessionId);
      _sessionStart = DateTime.now();
      sessionTask!
        ..reset()
        ..start();
      await addEvent(
          eventName: 'session_start',
          eventValue: '',
          dateTime: _sessionStart.format(con.dateFormat));
    }
  }

  Future<void> sessionEnd({String? endTime}) async {
    if (session != RabbitSession.start) return;
    sessionTask?.cancel();
    session = RabbitSession.end;
    await addEvent(eventName: 'session_end', eventValue: '', dateTime: endTime);
    session = RabbitSession.empty;
    _sessionId = '';
    SharedPreferences sharedPreference = await SharedPreferences.getInstance();
    await sharedPreference.remove('pauseTime');
    await sharedPreference.remove('sessionId');
    //await _exportJson();
  }

  Future<void> onPause() async {
    SharedPreferences sharedPreference = await SharedPreferences.getInstance();
    await sharedPreference.setString(
        'pauseTime', DateTime.now().toIso8601String());
    //print('save pauseTime ${DateTime.now().toIso8601String()}');

    sessionTask?.pause();
    resumeTask?.cancel();
  }

  Future<void> onResume() async {
    if (isFirstTime) return;

    SharedPreferences sharedPreference = await SharedPreferences.getInstance();
    var pauseTime = sharedPreference.getString('pauseTime') ?? '';

    if (pauseTime.isEmpty) return;

    var now = DateTime.now();
    var expireMilliseconds = _config.sessionAutoExpire * 60000;

    //print('pauseTime $pauseTime');
    sharedPreference.remove('pauseTime');

    var inactiveTime = now.difference(DateTime.parse(pauseTime));
    //print('${inactiveTime.inMilliseconds} > $expireMilliseconds');
    if (inactiveTime.inMilliseconds > expireMilliseconds) {
      //print('end start');
      _sessionId = sharedPreference.getString('sessionId') ??
          const Uuid().v4().toString();

      await sessionEnd(
          endTime: _sessionStart
              .add(Duration(minutes: _config.sessionAutoExpire))
              .format(con.dateFormat));
      await sessionStart();
    } else {
      //print('auto start');
      if (sessionTask != null && sessionTask!.isPaused) {
        var runTime = DateTime.parse(pauseTime).difference(_sessionStart);
        var sessionExpireMilliseconds = _config.sessionExpire * 60000;

        var leftTimeMilliseconds = sessionExpireMilliseconds -
            (runTime.inMilliseconds + inactiveTime.inMilliseconds);

        if (leftTimeMilliseconds < 0) {
          await sessionStart();
          return;
        }

        resumeTask = PausableTimer(Duration(milliseconds: leftTimeMilliseconds),
            () async {
          await _loopTime();
        });
        resumeTask!.start();
      }
    }

    await _eventRepo.deleteAfterRetentionDays(_config.retentionDays);
  }

  Future<void> addEvent(
      {required String eventName,
      required String eventValue,
      Map<dynamic, dynamic>? dataLayer,
      String? dateTime}) async {
    // refresh value
    await Future.wait<void>([
      _userInfoController.getData(),
      _locationController.getPosition(),
      _locationController.getLocale()
    ]);

    if (GetPlatform.isIOS &&
        _deviceInfoController.maId.isEmpty &&
        await Permission.appTrackingTransparency.isGranted) {
      await _deviceInfoController.getAdvertising();
    }

    var data = rabbitRequest
      ..memberId = _userInfoController.memberId
      ..ssoId = _userInfoController.ssoId
      ..sessionId = _sessionId
      ..sessionEvent = session
      ..sdkVersion = _versionController.sdkVersion
      ..eventName = eventName
      ..eventValue = eventValue
      ..osVer = _deviceInfoController.osVersion
      ..localeCountry = _locationController.localeCountry
      ..localeLang = _locationController.localeLang
      ..latitude = _locationController.position == null
          ? null
          : _checkData(
              _locationController.isGranted,
              _userInfoController.memberId,
              _config.consent,
              _locationController.position!.latitude.toString(),
              GetPlatform.isIOS && !_deviceInfoController.isPhysicalDevice
                  ? 'R00.00'
                  : '00.00')
      ..longitude = _locationController.position == null
          ? null
          : _checkData(
              _locationController.isGranted,
              _userInfoController.memberId,
              _config.consent,
              _locationController.position!.longitude.toString(),
              GetPlatform.isIOS && !_deviceInfoController.isPhysicalDevice
                  ? 'R00.00'
                  : '00.00')
      ..maid = _checkData(
          GetPlatform.isAndroid
              ? _config.permissionMaid
              : await _checkMaidIOSPermission(),
          _userInfoController.memberId,
          _config.consent,
          _deviceInfoController.maId.toString(),
          GetPlatform.isIOS && !_deviceInfoController.isPhysicalDevice
              ? 'R00000000-0000-0000-0000-000000000000'
              : '00000000-0000-0000-0000-000000000000')
      ..appVer = _versionController.appVer
      ..appVerNum = _versionController.appVerNum
      ..prevAppVer = _versionController.prevAppVer
      ..prevAppVerNum = _versionController.prevAppVerNum
      ..dataLayer = dataLayer
      ..time = dateTime ?? DateTime.now().format(con.dateFormat);

    await _conditionController.getStorageSize();
    // if storage < 1mb send to api
    if (_conditionController.currentFreeSpace < _config.storageMin) {
      var requests = <RabbitRequest>[];
      requests.add(data);
      await callAPI(requests);
      apiResult.add('storage = min');
      return;
    }
    // if storage between 1mb to 5mb delete oldest row
    else if (_conditionController.currentFreeSpace > _config.storageMin &&
        _conditionController.currentFreeSpace < _config.storageMax) {
      var data = await _eventRepo.first();
      if (data != null) {
        await _eventRepo.delete(data.id);
        apiResult.add('storage = deleteOld');
      }
    }

    // if free storage > 5mb insert normaly
    await _eventRepo.insert(data);
    changeDB.add('insert');

    // check collectionType = event : upload immediately
    if (_config.collectionType != DataCollectionType.event.name) {
      return;
    }

    if (!isLoading) await uploadEvent();
  }

  Future<List<RabbitRequest>> getEvent() async {
    return await _eventRepo.getAll();
  }

  Future<void> uploadEvent() async {
    var requests = await _eventRepo.getRequests();
    if (requests.isEmpty) return;

    // check cpu, memory, battery
    var validate = await _conditionController.checkSend();
    if (validate.containsValue(false)) {
      apiResult.add(
          'error [${validate.entries.where((element) => !element.value).map((e) => e.key).join(', ')}]');
      return;
    }

    await callAPI(requests);
  }

  Future<void> callAPI(List<RabbitRequest> data) async {
    // send data to api
    try {
      isLoading = true;
      final r = RetryOptions(
          delayFactor: const Duration(seconds: 3),
          maxAttempts: _config.retryTime);

      var response = await r.retry(() async {
        return await _api.post(
            con.eventBatchURL, data.map((e) => e.toJson()).toList());
      }, retryIf: (e) => e is SocketException || e is TimeoutException);

      // check offline mode not send
      if (_config.offlineMode) {
        response = false;
      }

      if (response) {
        apiResult.add('uploadEvent success');

        await _eventRepo.updateSend(data.map((e) => e.id).toList());

        // check deleteAfterSent
        if (_config.deleteAfterSent) {
          await _eventRepo.deleteAllKeys(data.map((e) => e.id).toList());
        }

        changeDB.add('refresh');

        isLoading = false;

        // check collectionType = event
        if (_config.collectionType != DataCollectionType.event.name) return;

        // loop self method
        await uploadEvent();
      } else {
        apiResult.add('uploadEvent fail');
      }
    } on DioError catch (e) {
      apiResult.add('uploadEvent fail : ${e.message.toString()}');
    } finally {
      isLoading = false;
    }
  }

  Future<void> clearCache() async {
    await _eventRepo.deleteAllSent();
    changeDB.add('deleteAllSent');
  }

  Future<void> clearAll() async {
    await _eventRepo.deleteAll();
    changeDB.add('deleteAll');
  }

  String _checkData(bool isPermission, String memberId, bool? isConsent,
      dynamic value, String defaultValue) {
    if (!isPermission) return '';

    if (isConsent == null) {
      // User: Anonymous
      if (memberId.isEmpty) {
        return value;
      }
      // User: Member
      else {
        return defaultValue;
      }
    }

    if (!isConsent) {
      return defaultValue;
    }

    return value;
  }

  Future<bool> _checkMaidIOSPermission() async {
    var version = _deviceInfoController.osVersion;
    var versionNum = int.parse(version.split('.').first);
    if (versionNum > 13) {
      return await Permission.appTrackingTransparency.isGranted;
    } else {
      return _config.permissionMaid;
    }
  }
}
